@@ -1,8 +1,8 @@
-/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
+/*
 
   digest -- hash digest functions for R
 
-  Copyright (C) 2003 - 2016  Dirk Eddelbuettel <edd@debian.org>
+  Copyright (C) 2003 - 2023  Dirk Eddelbuettel <edd@debian.org>
 
   This file is part of digest.
 
@@ -37,6 +37,7 @@
 #include "xxhash.h"
 #include "pmurhash.h"
 #include "blake3.h"
+#include "crc32c.h"
 
 #ifdef _WIN32
 #include <Windows.h>
@@ -124,7 +125,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for (j = 0; j < 16; j++)
-                sprintf(output + j * 2, "%02x", md5sum[j]);
+                snprintf(output + j * 2, 3, "%02x", md5sum[j]);
 
         break;
     }
@@ -141,7 +142,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for ( j = 0; j < 20; j++ )
-                sprintf( output + j * 2, "%02x", sha1sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha1sum[j] );
 
         break;
     }
@@ -152,7 +153,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         val  = digest_crc32(0L, 0, 0);
         val  = digest_crc32(val, (unsigned char*) txt, (unsigned) l);
 
-        sprintf(output, "%08x", (unsigned int) val);
+        snprintf(output, 128, "%08x", (unsigned int) val);
         break;
     }
     case 4: {     /* sha256 case */
@@ -168,7 +169,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
 
         if (!leaveRaw)
             for ( j = 0; j < 32; j++ )
-                sprintf( output + j * 2, "%02x", sha256sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha256sum[j] );
 
         break;
     }
@@ -198,21 +199,21 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     }
     case 6: {     /* xxhash32 case */
         unsigned int val =  XXH32(txt, nChar, seed);
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 7: {     /* xxhash64 case */
         unsigned long long val =  XXH64(txt, nChar, seed);
 #if defined(WIN32) && !defined(_UCRT)
-        sprintf(output, "%016" PRIx64, val);
+        snprintf(output, 128, "%016" PRIx64, val);
 #else
-        sprintf(output, "%016llx", val);
+        snprintf(output, 128, "%016llx", val);
 #endif
         break;
     }
     case 8: {     /* MurmurHash3 32 */
         unsigned int val = PMurHash32(seed, txt, nChar);
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 10: {     /* blake3 */
@@ -226,11 +227,17 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             memcpy(output, val, BLAKE3_OUT_LEN);
         } else {
             for (size_t i = 0; i < BLAKE3_OUT_LEN; i++) {
-                sprintf(output + i * 2, "%02x", val[i]);
+                snprintf(output + i * 2, 3, "%02x", val[i]);
             }
         }
         break;
     }
+    case 11: {		/* crc32c */
+        uint32_t crc = 0;       /* initial value, can be zero */
+        crc = crc32c_extend(crc, (const uint8_t*) txt, (size_t) nChar);
+        snprintf(output, 128, "%08x", crc);
+        break;
+    }
     case 101: {     /* md5 file case */
         int j;
         md5_context ctx;
@@ -255,7 +262,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         memcpy(output, md5sum, 16);
         if (!leaveRaw)
             for (j = 0; j < 16; j++)
-                sprintf(output + j * 2, "%02x", md5sum[j]);
+                snprintf(output + j * 2, 3, "%02x", md5sum[j]);
         break;
     }
     case 102: {     /* sha1 file case */
@@ -281,7 +288,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         memcpy(output, sha1sum, 20);
         if (!leaveRaw)
             for ( j = 0; j < 20; j++ )
-                sprintf( output + j * 2, "%02x", sha1sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha1sum[j] );
         break;
     }
     case 103: {     /* crc32 file case */
@@ -300,7 +307,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
                 val  = digest_crc32(val , buf, (unsigned) nChar);
         }
-        sprintf(output, "%08x", (unsigned int) val);
+        snprintf(output, 128, "%08x", (unsigned int) val);
         break;
     }
     case 104: {     /* sha256 file case */
@@ -326,7 +333,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         memcpy(output, sha256sum, 32);
         if (!leaveRaw)
             for ( j = 0; j < 32; j++ )
-                sprintf( output + j * 2, "%02x", sha256sum[j] );
+                snprintf( output + j * 2, 3, "%02x", sha256sum[j] );
         break;
     }
     case 105: {     /* sha2-512 file case */
@@ -396,7 +403,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         unsigned int val =  XXH32_digest(state);
         XXH32_freeState(state);
 
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 107: {     /* xxhash64 */
@@ -429,9 +436,9 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         XXH64_freeState(state);
 
 #ifdef WIN32
-        sprintf(output, "%016" PRIx64, val);
+        snprintf(output, 128, "%016" PRIx64, val);
 #else
-        sprintf(output, "%016llx", val);
+        snprintf(output, 128, "%016llx", val);
 #endif
         break;
     }
@@ -457,7 +464,7 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         }
         unsigned int val = PMurHash32_Result(h1, carry, total_length);
 
-        sprintf(output, "%08x", val);
+        snprintf(output, 128, "%08x", val);
         break;
     }
     case 110: {     /* blake3 file case */
@@ -483,11 +490,29 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
             memcpy(output, val, BLAKE3_OUT_LEN);
         } else {
             for (size_t i = 0; i < BLAKE3_OUT_LEN; i++) {
-                sprintf(output + i * 2, "%02x", val[i]);
+                snprintf(output + i * 2, 3, "%02x", val[i]);
             }
         }
         break;
     }
+    case 111: {		/* crc32c */
+        unsigned char buf[BUF_SIZE];
+        uint32_t crc = 0;
+
+        if (skip > 0) fseek(fp, skip, SEEK_SET);
+        if (length>=0) {
+            while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0 && length>0) {
+                if (nChar>length) nChar=length;
+                crc = crc32c_extend(crc, (const uint8_t*) buf, (size_t) nChar);
+                length -= nChar;
+            }
+        } else {
+            while ( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0)
+                crc = crc32c_extend(crc, (const uint8_t*) buf, (size_t) nChar);
+        }
+        snprintf(output, 128, "%08x", (unsigned int) crc);
+        break;
+    }
     default: {
         error("Unsupported algorithm code"); /* should not be reached due to test in R */ /* #nocov */
     }
